#include "api.h"

#include <stddef.h>

// taken from poly1305.h from OpenSSL

#define POLY1305_BLOCK_SIZE  16
#define POLY1305_DIGEST_SIZE 16
#define POLY1305_KEY_SIZE    32

typedef struct poly1305_context POLY1305;

typedef void (*poly1305_blocks_f) (void *ctx, const unsigned char *inp,
                                   size_t len, unsigned int padbit);
typedef void (*poly1305_emit_f) (void *ctx, unsigned char mac[16],
                                 const unsigned int nonce[4]);

struct poly1305_context {
    double opaque[24];  /* large enough to hold internal state, declared
                         * 'double' to ensure at least 64-bit invariant
                         * alignment across all platforms and
                         * configurations */
    unsigned int nonce[4];
    unsigned char data[POLY1305_BLOCK_SIZE];
    size_t num;
    struct {
        poly1305_blocks_f blocks;
        poly1305_emit_f emit;
    } func;
};

// taken from SUPERCOP

int crypto_verify_16(const unsigned char *x,const unsigned char *y)
{
  unsigned int differentbits = 0;
#define F(i) differentbits |= x[i] ^ y[i];
  F(0)
  F(1)
  F(2)
  F(3)
  F(4)
  F(5)
  F(6)
  F(7)
  F(8)
  F(9)
  F(10)
  F(11)
  F(12)
  F(13)
  F(14)
  F(15)
  return (1 & ((differentbits - 1) >> 8)) - 1;
#undef F
}

//

extern void Poly1305_Init(POLY1305*, const uint8_t *);
extern void Poly1305_Update(POLY1305*, const uint8_t *, uint64_t);
extern void Poly1305_Final(POLY1305*, uint8_t*);

int external_onetimeauth_poly1305_amd64_avx2(
 uint8_t *mac,
 const uint8_t *input,
 uint64_t input_length,
 const uint8_t *key
)
{
  POLY1305 poly1305;
  Poly1305_Init(&poly1305, key);
  Poly1305_Update(&poly1305, input, input_length);
  Poly1305_Final(&poly1305, mac);
  return 0;
}

int external_onetimeauth_poly1305_amd64_avx2_verify(
 const uint8_t *mac,
 const uint8_t *input,
 uint64_t input_length,
 const uint8_t *key
)
{
  unsigned char correct[16];
  external_onetimeauth_poly1305_amd64_avx2(correct,input,input_length,key);
  return crypto_verify_16(mac,correct); // not to be benchmarked
}

